package ru.dbotthepony.mc.otm.client.screen.panels

import com.mojang.blaze3d.systems.RenderSystem
import com.mojang.blaze3d.vertex.PoseStack
import net.minecraft.client.Minecraft
import net.minecraft.client.gui.narration.NarratableEntry
import net.minecraft.client.gui.narration.NarratableEntry.NarrationPriority
import net.minecraft.client.gui.narration.NarrationElementOutput
import net.minecraft.client.resources.sounds.SimpleSoundInstance
import net.minecraft.network.chat.Component
import net.minecraft.sounds.SoundEvents
import org.lwjgl.opengl.GL30
import ru.dbotthepony.mc.otm.client.render.*
import ru.dbotthepony.mc.otm.client.screen.MatteryScreen

open class FramePanel(
	screen: MatteryScreen<*>,
	parent: EditablePanel?,
	x: Float,
	y: Float,
	width: Float,
	height: Float,
	protected var title: Component
) : EditablePanel(screen, parent, x, y, width, height), NarratableEntry {
	constructor(screen: MatteryScreen<*>, width: Float, height: Float, title: Component) : this(screen, null, 0f, 0f, width, height, title)

	enum class Position(val width: Float, val height: Float, val active_width: Float, val active_height: Float) {
		TOP(28f, 28f, 28f, 32f),
		// TODO: а оно вообще нужно?
		LEFT(28f, 28f, 32f, 28f),
		RIGHT(28f, 28f, 32f, 28f),
		BOTTOM(28f, 28f, 28f, 32f);
	}

	open inner class FrameTabPanel(position: Position, on_open: Runnable? = null, on_close: Runnable? = null) :
		EditablePanel(
			this@FramePanel.screen,
			this@FramePanel,
			0f,
			0f,
			position.width,
			position.height
		) {

		val frameTabPosition: Position = position
		var isActive = false
		var initial = false

		var onOpen: Runnable? = on_open
		var onClose: Runnable? = on_close

		open fun bindOnOpen(value: Runnable?) {
			onOpen = value
		}

		open fun bindOnClose(value: Runnable?) {
			onClose = value
		}

		fun showHidePanels(input: List<EditablePanel>) {
			onOpen = Runnable {
				for (child in input) {
					child.visible = true
				}
			}

			onClose = Runnable {
				for (child in input) {
					child.visible = false
				}
			}
		}

		override fun innerRender(stack: PoseStack, mouseX: Float, mouseY: Float, partialTick: Float) {
			if (frameTabPosition == Position.TOP) {
				val width: Float
				val height: Float

				if (isActive) {
					width = frameTabPosition.active_width
					height = frameTabPosition.active_height
					RECTANGLE.middle.render(stack, 2f, 2f, width - 4, height - 2)
				} else {
					width = frameTabPosition.width
					height = frameTabPosition.height
					tabBackground.render(stack, 2f, 2f, width - 4, height - 2)
				}

				RECTANGLE.top.renderW(stack, 3f, 0f, width - 6)
				RECTANGLE.left.renderH(stack, 0f, 3f, (height - if (isActive) if (initial) 2 else 4 else 3))

				RECTANGLE.right.renderH(stack, width - RECTANGLE.right.w, 3f, (height - if (isActive) 4 else 3))

				RECTANGLE.topLeft.render(stack, 0f, 0f)
				RECTANGLE.topRight.render(stack, width - RECTANGLE.topRight.w, 0f)

				if (isActive) {
					if (!initial) {
						tabLeftConnection.render(stack, 0f, height - tabLeftConnection.h - 1)
					}

					tabRightConnection.render(
						stack,
						width - tabRightConnection.w,
						height - tabLeftConnection.h - 1
					)
				}
			}
		}

		protected open fun onOpen() {
			onOpen?.run()
		}

		protected open fun onClose() {
			onClose?.run()
		}

		override fun mouseClickedInner(x: Double, y: Double, flag: Int): Boolean {
			if (isActive) {
				return true
			}

			for (tab in tabsTop) {
				if (tab !== this) {
					if (tab.isActive) {
						tab.onClose()
						tab.isActive = false
					}
				}
			}

			for (tab in tabsLeft) {
				if (tab !== this) {
					if (tab.isActive) {
						tab.onClose()
						tab.isActive = false
					}
				}
			}

			for (tab in tabsRight) {
				if (tab !== this) {
					if (tab.isActive) {
						tab.onClose()
						tab.isActive = false
					}
				}
			}

			for (tab in tabsBottom) {
				if (tab !== this) {
					if (tab.isActive) {
						tab.onClose()
						tab.isActive = false
					}
				}
			}

			isActive = true
			onOpen()
			Minecraft.getInstance().soundManager.play(SimpleSoundInstance.forUI(SoundEvents.UI_BUTTON_CLICK, 1.0f))

			return true
		}
	}

	protected val tabsTop = ArrayList<FrameTabPanel>()
	protected val tabsLeft = ArrayList<FrameTabPanel>()
	protected val tabsRight = ArrayList<FrameTabPanel>()
	protected val tabsBottom = ArrayList<FrameTabPanel>()

	@JvmOverloads
	fun addTab(position: Position, on_open: Runnable? = null, on_close: Runnable? = null): FrameTabPanel {
		val tab = FrameTabPanel(position, on_open, on_close)
		doAddTab(tab)
		return tab
	}

	protected fun doAddTab(tab: FrameTabPanel) {
		if (tabsTop.size == 0 && tabsLeft.size == 0 && tabsRight.size == 0 && tabsBottom.size == 0) {
			tab.isActive = true
		}

		when (tab.frameTabPosition) {
			Position.TOP -> tabsTop.add(tab)
			Position.LEFT -> tabsLeft.add(tab)
			Position.RIGHT -> tabsRight.add(tab)
			Position.BOTTOM -> tabsBottom.add(tab)
		}
	}

	override fun performLayout() {
		for (i in tabsTop.indices) {
			val tab = tabsTop[i]
			tab.setPos(i * Position.TOP.width, -Position.TOP.height)
			tab.initial = i == 0
		}

		super.performLayout()
	}

	protected var dragging = false

	init {
		setDockPadding(PADDING, PADDING_TOP, PADDING, PADDING)
	}

	override fun mouseClickedInner(x: Double, y: Double, button: Int): Boolean {
		val (_, y1) = screenToLocal(x, y)

		if (parent == null && y1 >= 0 && y1 <= 12) {
			trapMouseInput = true
			dragging = true
			return true
		}

		return super.mouseClickedInner(x, y, button)
	}

	override fun mouseReleasedInner(x: Double, y: Double, button: Int): Boolean {
		if (dragging) {
			trapMouseInput = false
			dragging = false
			return true
		}

		return super.mouseReleasedInner(x, y, button)
	}

	override fun mouseDraggedInner(x: Double, y: Double, button: Int, xDelta: Double, yDelta: Double): Boolean {
		if (dragging) {
			this.x += xDelta.toFloat()
			this.y += yDelta.toFloat()
		}

		return super.mouseDraggedInner(x, y, button, xDelta, yDelta)
	}

	override fun innerRender(stack: PoseStack, mouseX: Float, mouseY: Float, partialTick: Float) {
		RECTANGLE.render(stack, width = width, height = height)

		// title
		RenderSystem.depthFunc(GL30.GL_ALWAYS)
		screen.font.draw(stack, title, 8f, 5f, 4210752)
		RenderSystem.depthFunc(GL30.GL_ALWAYS)
	}

	override fun narrationPriority(): NarrationPriority {
		return NarrationPriority.NONE
	}

	override fun updateNarration(narrationElementOutput: NarrationElementOutput) {}

	fun toScreenCenter() {
		setPos((screen.width / 2 - (width / 2).toInt()).toFloat(), (screen.height / 2 - (height / 2).toInt()).toFloat())
	}

	companion object {
		const val PADDING = 8f
		const val PADDING_TOP = 14f

		val RECTANGLE = StretchingRectangleElement(
			topLeft = WidgetLocation.WIDGETS.element(x = 18f, y = 0f, w = 6f, h = 6f),
			topRight = WidgetLocation.WIDGETS.element(x = 24f, y = 0f, w = 6f, h = 6f),
			left = WidgetLocation.WIDGETS.element(x = 18f, y = 4f, w = 3f, h = 5f),
			right = WidgetLocation.WIDGETS.element(x = 25f, y = 3f, w = 5f, h = 5f),
			top = WidgetLocation.WIDGETS.element(x = 22f, y = 0f, w = 5f, h = 3f),
			bottomLeft = WidgetLocation.WIDGETS.element(x = 18f, y = 6f, w = 6f, h = 6f),
			bottomRight = WidgetLocation.WIDGETS.element(x = 24f, y = 6f, w = 6f, h = 6f),
			bottom = WidgetLocation.WIDGETS.element(x = 21f, y = 9f, w = 5f, h = 3f),
			middle = WidgetLocation.WIDGETS.element(x = 30f, y = 12f, w = 6f, h = 6f),
			padding = DockProperty(-3f, -3f, -3f, -3f)
		)

		val tabRightConnection 				= WidgetLocation.WIDGETS.element(x = 30f, y = 0f, w = 3f, h = 5f)
		val tabLeftConnection 				= WidgetLocation.WIDGETS.element(x = 33f, y = 0f, w = 3f, h = 5f)
		val tabBackground 					= WidgetLocation.WIDGETS.element(x = 30f, y = 6f, w = 6f, h = 6f)
	}
}
